# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools as it
from montepy.constants import ASCII_CEILING
from montepy.utilities import *
import os


class MCNP_InputFile:
    """
    A class to represent a distinct input file.

    .. Note::
        this is a bare bones implementation to be fleshed out in the future.

    .. versionchanged:: 0.3.0
        Added the overwrite attribute.

    :param path: the path to the input file
    :type path: str
    :param parent_file: the parent file for this file if any. This occurs when a "read" input is used.
    :type parent_file: str
    :param overwrite: Whether to overwrite the file 'path' if it exists
    :type overwrite: bool
    """

    def __init__(self, path, parent_file=None, overwrite=False):
        self._path = path
        self._parent_file = parent_file
        self._lineno = 1
        self._replace_with_space = False
        self._overwrite = overwrite
        self._mode = None
        self._fh = None

    @make_prop_pointer("_path")
    def path(self):
        """
        The path for the file.

        :rtype: str
        """
        pass

    @property
    def name(self):
        return self.path

    @make_prop_pointer("_parent_file")
    def parent_file(self):
        """
        The parent file for this file.

        This is only used when this file is pointed to by a "read" input.

        :rtype: str
        """
        pass

    @make_prop_pointer("_lineno")
    def lineno(self):
        """
        The current line number being read in the file.

        This is 1-indexed.

        :rtype: int
        """
        pass

    def open(self, mode, encoding="ascii", replace=True):
        """
        Opens the underlying file, and returns self.

        This should only ever be completed from within a ``with`` statement.
        For this reason, a ``close`` functional is intentionally
        not provided.


        .. Note::
            For different encoding schemes see the available list
            `here <https://docs.python.org/3.9/library/codecs.html#standard-encodings>`_.

            CP1252 is commonly referred to as "extended-ASCII".
            You may have success with this encoding for working with special characters.

        .. versionchanged:: 0.2.11
            Added guardrails to raise FileExistsError and IsADirectoryError.

        :param mode: the mode to open the file in
        :type mode: str
        :param encoding: The encoding scheme to use. If replace is true, this is ignored, and changed to ASCII
        :type encoding: str
        :param replace: replace all non-ASCII characters with a space (0x20)
        :type replace: bool
        :returns: self
        :raises FileExistsError: if a file already exists with the same path while writing.
        :raises IsADirectoryError: if the path given is actually a directory while writing.
        """
        if "r" in mode:
            if replace:
                self._replace_with_space = True
                mode = "rb"
                encoding = None
        self._mode = mode
        if "w" in mode:
            if os.path.isfile(self.path) and self._overwrite is not True:
                raise FileExistsError(
                    f"{self.path} already exists, and overwrite is not set."
                )
            if os.path.isdir(self.path):
                raise IsADirectoryError(
                    f"{self.path} is a directory, and cannot be overwritten."
                )
        self._fh = open(self.path, mode, encoding=encoding)
        return self

    def __enter__(self):
        self._fh.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = self._fh.__exit__(exc_type, exc_val, exc_tb)
        self._fh = None
        return status

    def __iter__(self):
        for lineno, line in enumerate(self._fh):
            self._lineno = lineno + 1
            if self._mode == "rb" and self._replace_with_space:
                line = self._clean_line(line)
            yield line

    @staticmethod
    def _clean_line(line):
        new_line = bytes([code if code < ASCII_CEILING else ord(" ") for code in line])
        line = new_line.decode("ascii")
        line = line.replace("\r\n", "\n").replace("\r", "\n")
        return line

    def read(self, size=-1):
        """ """
        if self._fh:
            ret = self._fh.read(size)
            if self._mode == "rb" and self._replace_with_space:
                ret = self._clean_line(ret)
            self._lineno += ret.count("\n")
            return ret

    def readline(self, size=-1):
        """ """
        if self._fh:
            ret = self._fh.readline(size)
            if self._mode == "rb" and self._replace_with_space:
                ret = self._clean_line(ret)
            self._lineno += ret.count("\n")
            return ret

    def write(self, to_write):
        if self._fh:
            self._lineno += to_write.count("\n")
            return self._fh.write(to_write)

    def __str__(self):
        return self.name



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.cells import Cells
from montepy.input_parser.mcnp_input import Input
from montepy.input_parser.block_type import BlockType
from montepy.input_parser import syntax_node
from montepy.numbered_mcnp_object import Numbered_MCNP_Object


class Universe(Numbered_MCNP_Object):
    """
    Class to represent an MCNP universe, but not handle the input
    directly.

    :param number: The number for the universe, must be ≥ 0
    :type number: int
    """

    def __init__(self, number):
        if not isinstance(number, int):
            raise TypeError("number must be int")
        if number < 0:
            raise ValueError(f"Universe number must be ≥ 0. {number} given.")
        self._number = number

        class Parser:
            def parse(self, token_gen, input):
                return syntax_node.SyntaxNode("fake universe", {})

        super().__init__(Input(["U"], BlockType.DATA), Parser())

    @property
    def cells(self):
        """
        A generator of the cell objects in this universe.

        :return: a generator returning every cell in this universe.
        :rtype: Generator
        """
        if self._problem:
            for cell in self._problem.cells:
                if cell.universe == self:
                    yield cell

    def claim(self, cells):
        """
        Take the given cells and move them into this universe, and out of their original universe.

        Can be given a single Cell, a list of cells, or a Cells object.

        :param cells: the cell(s) to be claimed
        :type cells: Cell, list, or Cells
        :raises TypeError: if bad parameter is given.
        """
        if not isinstance(cells, (montepy.Cell, list, Cells)):
            raise TypeError(f"Cells being claimed must be a Cell, list, or Cells")
        if isinstance(cells, list):
            cells = Cells(cells)
        if isinstance(cells, montepy.Cell):
            cells = Cells([cells])

        for cell in cells:
            cell.universe = self

    @property
    def number(self):
        """
        The number for this Universe.

        Universe 0 is protected, and a Universe cannot be set 0,
        if it is not already Universe 0.
        """
        return self._number

    @number.setter
    def number(self, number):
        if not isinstance(number, int):
            raise TypeError("number must be int")
        if number <= 0:
            raise ValueError("Universe number must be > 0")
        if self._problem:
            self._problem.universes.check_number(number)
        self._number = number

    @property
    def old_number(self):
        """
        Original universe number from the input file.
        """
        return self._number

    def _update_values(self):
        pass

    def __str__(self):
        return f"Universe({self.number})"

    def __repr__(self):
        return (
            f"Universe: Number: {self.number} "
            f"Problem: {'set' if self._problem else 'not set'}, "
            f"Cells: {[cell.number for cell in self.cells] if self._problem else ''}"
        )
